package com.tzld.piaoquan.recommend.model

import ml.dmlc.xgboost4j.scala.spark.XGBoostClassifier
import org.apache.commons.lang.math.NumberUtils
import org.apache.commons.lang3.StringUtils
import org.apache.hadoop.io.compress.GzipCodec
import org.apache.spark.ml.evaluation.BinaryClassificationEvaluator
import org.apache.spark.ml.feature.VectorAssembler
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.types.{DataTypes, StructField}
import org.apache.spark.sql.{Dataset, Row, RowFactory, SparkSession}

import java.util
import scala.io.Source

object train_01_xgb_ad_20240808{
  def main(args: Array[String]): Unit = {
    val spark = SparkSession
      .builder()
      .appName(this.getClass.getName)
      .getOrCreate()
    val sc = spark.sparkContext
//    val features = Array("cpa", "b2_12h_ctr", "b2_12h_ctcvr", "b2_12h_cvr", "b2_12h_conver", "b2_12h_click", "b2_12h_conver*log(view)", "b2_12h_conver*ctcvr", "b2_7d_ctr", "b2_7d_ctcvr", "b2_7d_cvr", "b2_7d_conver", "b2_7d_click", "b2_7d_conver*log(view)", "b2_7d_conver*ctcvr")

    val loader = getClass.getClassLoader
    val resourceUrl = loader.getResource("20240703_ad_feature_name.txt")
    val content =
      if (resourceUrl != null) {
        val content = Source.fromURL(resourceUrl).getLines().mkString("\n")
        Source.fromURL(resourceUrl).close()
        content
      } else {
        ""
      }
    println(content)
    val features = content.split("\n")
      .map(r => r.replace(" ", "").replaceAll("\n", ""))
      .filter(r => r.nonEmpty)


    val trainData = createData(
      sc.textFile("/dw/recommend/model/33_ad_train_data_v4/20240724"),
      features
    )
    println("train data size：" + trainData.count())

    val fields = Array(
      DataTypes.createStructField("label", DataTypes.IntegerType, true)
    ) ++ features.map(f => DataTypes.createStructField(f, DataTypes.DoubleType, true))
    val schema = DataTypes.createStructType(fields)
    val trainDataSet: Dataset[Row] = spark.createDataFrame(trainData, schema)
    val vectorAssembler = new VectorAssembler().setInputCols(features).setOutputCol("features")
    val xgbInput = vectorAssembler.transform(trainDataSet).select("features","label")
//    val xgbParam = Map("eta" -> 0.01f,
//      "max_depth" -> 5,
//      "objective" -> "binary:logistic",
//      "num_class" -> 3)
    val xgbClassifier = new XGBoostClassifier()
      .setEta(0.01f)
      .setMissing(0.0f)
      .setMaxDepth(5)
      .setNumRound(1000)
      .setSubsample(0.8)
      .setColsampleBytree(0.8)
      .setScalePosWeight(1)
      .setObjective("binary:logistic")
      .setEvalMetric("auc")
      .setFeaturesCol("features")
      .setLabelCol("label")
      .setNthread(1)
      .setNumWorkers(22)
    val model = xgbClassifier.fit(xgbInput)


    val testData = createData(
      sc.textFile("/dw/recommend/model/33_ad_train_data_v4/20240725"),
      features
    )
    val testDataSet = spark.createDataFrame(testData, schema)
    val testDataSetTrans = vectorAssembler.transform(testDataSet).select("features","label")
    val predictions = model.transform(testDataSetTrans)

    val saveData = predictions.select("label", "prediction", "features", "rawPrediction", "probability").rdd
      .map(r =>{
        (r.get(0), r.get(1), r.get(2), r.get(3), r.get(4)).productIterator.mkString("\t")
    })
    val hdfsPath = "/dw/recommend/model/checkpoint_xgbtest"
    if (hdfsPath.nonEmpty && hdfsPath.startsWith("/dw/recommend/model/")) {
      println("删除路径并开始数据写入:" + hdfsPath)
      MyHdfsUtils.delete_hdfs_path(hdfsPath)
      saveData.repartition(100).saveAsTextFile(hdfsPath, classOf[GzipCodec])
    } else {
      println("路径不合法，无法写入:" + hdfsPath)
    }



    val evaluator = new BinaryClassificationEvaluator()
      .setLabelCol("label")
      .setRawPredictionCol("probability")
      .setMetricName("areaUnderROC")
    val auc = evaluator.evaluate(predictions.select("label", "probability"))
    println("zhangbo:auc:" + auc)
  }


  def createData(data: RDD[String], features: Array[String]): RDD[Row] = {
    data.map(r => {
      val line: Array[String] = StringUtils.split(r, '\t')
      val label: Int = NumberUtils.toInt(line(0))
      val map: util.Map[String, Double] = new util.HashMap[String, Double]
      for (i <- 1 until line.length) {
        val fv: Array[String] = StringUtils.split(line(i), ':')
        map.put(fv(0), NumberUtils.toDouble(fv(1), 0.0))
      }

      val v: Array[Any] = new Array[Any](features.length + 1)
      v(0) = label
      for (i <- 0 until features.length) {
        v(i + 1) = map.getOrDefault(features(i), 0.0d)
      }
      Row(v: _*)
    })
  }
}



//rabit_timeout -> -1
//scale_pos_weight -> 1.0
//seed -> 0
//handle_invalid -> error
//features_col -> features
//label_col -> label
//num_workers -> 1
//subsample -> 0.8
//max_depth -> 5
//probability_col -> probability
//raw_prediction_col -> rawPrediction
//tree_limit -> 0
//dmlc_worker_connect_retry -> 5
//train_test_ratio -> 1.0
//use_external_memory -> false
//objective -> binary:logistic
//eval_metric -> auc
//num_round -> 1000
//missing -> 0.0
//rabit_ring_reduce_threshold -> 32768
//tracker_conf -> TrackerConf(0,python,,)
//eta -> 0.009999999776482582
//colsample_bytree -> 0.8
//allow_non_zero_for_missing -> false
//nthread -> 8
//prediction_col -> prediction